{
  open Parser        (* The type token is defined in parser.mli *)
  open Lexing
  exception Eof

  let comment_count = ref 0
}

rule token = parse
  | [' ' '\t']          { token lexbuf }
  | ['\n']              { lexbuf.lex_curr_p <- { lexbuf.lex_curr_p with pos_lnum = lexbuf.lex_curr_p.pos_lnum + 1 }; token lexbuf }
  | "token"             { TOKEN }
  | "prec"              { PREC }
  | "::="               { COLONCOLONEQUAL }
  | ":"                 { COLON }
  | "."                 { DOT }
  | ","                 { COMMA }
  | "("                 { LEFTPAR }
  | ")"                 { RIGHTPAR }
  | "->"                { ARROW }
  | "=>"                { DOUBLEARROW }
  | "|"                 { VERT }
  | ";"                 { SEMI }
  | "#"                 { SHARP }
  | "*"                 { STAR }
  | "+"                 { PLUS }
  | "?"                 { QUESTIONMARK }
  | "!"                 { BANG }
  | "{"                 { LEFTACC }
  | "}"                 { RIGHTACC }
  | ['0'-'9']+          { INT(int_of_string(Lexing.lexeme lexbuf)) }
  | ['a'-'z' 'A'-'Z'] ['a'-'z' 'A'-'Z' '_' '0'-'9']*
                        { ID(Lexing.lexeme lexbuf) }
  | eof                 { EOF }
  | "(*"                { comment_count := 1; if comment lexbuf then token lexbuf else EOF }
  | "(*)"               { token lexbuf }
  | "*)"                { token lexbuf }

and comment = parse
  | "(*"                { comment_count := !comment_count + 1; comment lexbuf }
  | "(*)"               { comment lexbuf }
  | "*)"                { comment_count := !comment_count - 1; if !comment_count <= 0 then true else comment lexbuf }
  | eof                 { false }
  | _                   { comment lexbuf }
