%{
  open Parsetree

  let prec_of_string = function
    "left" -> PLeft
  | "right" -> PRight
  | "nonassoc" -> PNonAssoc
  | s -> raise (Parse_error_string ("Unknown precedence: "^s^". Valid precedences are left, right, and nonassoc."))
%}

%token COLONCOLONEQUAL
%token COLON
%token DOT
%token COMMA
%token LEFTPAR
%token RIGHTPAR
%token ARROW
%token DOUBLEARROW
%token VERT
%token SEMI
%token SHARP
%token STAR
%token PLUS
%token QUESTIONMARK
%token LEFTACC
%token RIGHTACC
%token <int> INT
%token <string> ID
%token EOF
%token TOKEN
%token PREC
%token BANG

%left VERT
%left COLON
%left COMMA
%left BANG
%nonassoc SHARP
%nonassoc STAR PLUS QUESTIONMARK

%start grammar
%type <Parsetree.grammar> grammar

%%

grammar:
  ID COLONCOLONEQUAL tok grammar
  { (GRule($1, $3))::$4 }
| TOKEN ID COLON ID grammar
  { (GToken($2, $4))::$5 }
| PREC prec_list grammar
  { (GPrec $2)::$3 }
| EOF
  { [] }
;

prec_list:
  prec COMMA prec_list
  { $1@$3 }
| prec
  { $1 }
;

prec:
  id_plus COLON ID
  { List.map (fun id -> (id, prec_of_string $3)) $1 }
;

id_plus:
  ID id_plus
  { $1::$2 }
| ID
  { [$1] }
;

tok:
  ID
  { TIdent $1 }
| SHARP tok
  { TIgnore $2 }
| LEFTPAR tok RIGHTPAR
  { $2 }
| tok COMMA tok
  { TSeq($1, $3) }
| tok VERT tok
  { TBranch($1, $3) }
| tok STAR
  { TStar $1 }
| tok PLUS
  { TPlus $1 }
| tok QUESTIONMARK
  { TOption $1 }
| tok BANG tok
  { TBang($1, $3) }
| ID COLON tok
  { TLabeled($1, $3) }
| LEFTACC tok RIGHTACC
  { TSub $2 }
| PREC ID
  { TPrec $2 }
;

%%
