open Parsetree
open Yacc
open Typetree

exception TypeUnionError of compiled_type * compiled_type

let rec find_label label = function
  [] -> None
| (l, t)::_ when label = l -> Some t
| _::tl -> find_label label tl

let rec variant_union (v: (string * compiled_type) list) = function
  [] -> v
| (l, t)::tl -> (
    match find_label l v with
      None -> variant_union ((l, t)::v) tl
    | Some u -> variant_union ((l, type_union None (t, u))::v) tl
  )

and type_union rulename = function
  x, y when x = y -> x
| CTVariant v1, CTVariant v2 -> CTVariant(variant_union v1 v2)
(*| (CTIdent n, x
| x, CTIdent n) when rulename = Some n -> x*)
| x, y -> raise (TypeUnionError(x, y))

let simplify_tuple l =
  match List.filter (fun t -> t <> CTUnit) l with
    [] -> CTUnit
  | [t] -> t
  | lf -> CTTuple(lf, false)

let rec compile_tok (tt: TTSet.t) rulename = function
  TIdent s -> (
    match TTSet.fold (fun (name, t) a -> if name = s then Some t else a) tt None with
      Some Some t -> CTIdent t
    | Some None -> CTUnit
    | None -> CTIdent s
  )
| TIgnore _ -> CTUnit
| TSeq(t, u) -> simplify_tuple (
    match compile_tok tt None t, compile_tok tt None u with
      CTTuple(l1, false), CTTuple(l2, false) -> l1@l2
    | t1, CTTuple(l2, false) -> t1::l2
    | CTTuple(l1, false), t2 -> l1@[t2]
    | t1, t2 -> [t1; t2]
  )
| TBranch(t, u) -> type_union rulename (compile_tok tt rulename t, compile_tok tt rulename u)
| TStar t
| TPlus t
| TBang(t, _) -> (
    match compile_tok tt None t with
      CTUnit -> CTUnit
    | u -> CTList u
  )
| TOption t -> CTOption (compile_tok tt None t)
  (*
    match compile_tok tt t with
      CTUnit -> CTUnit
    | u -> CTOption u
  *)
| TLabeled(l, t) -> CTVariant([l, compile_tok tt None t])
| TSub t -> (
    match compile_tok tt rulename t with
      CTTuple(l, false) -> CTTuple(l, true)
    | x -> CTTuple([x], true)
  )
| TPrec _ -> CTUnit

let rec compile_grammar tt aux = function
  [] -> aux
| (GRule(name, tok))::tl -> compile_grammar tt ((name, compile_tok tt (Some name) tok)::aux) tl
| _::tl -> compile_grammar tt aux tl

let compile_grammar tt = compile_grammar tt []
