type compiled_type =
  CTIdent of string
| CTList of compiled_type
| CTOption of compiled_type
| CTVariant of (string * compiled_type) list
| CTUnit
| CTTuple of compiled_type list * bool (* bool = true pour le cas TSub *)

type compiled_types = string * compiled_type list

let rec compiled_type_to_buf buf = function
  CTIdent s -> Buffer.add_string buf s
| CTList t ->
    compiled_type_to_buf buf t;
    Buffer.add_string buf " list"
| CTOption t ->
    compiled_type_to_buf buf t;
    Buffer.add_string buf " option"
| CTVariant v ->
    Buffer.add_string buf " [ ";
    let vert = ref false in
    List.iter (fun (l, t) ->
      if !vert then Buffer.add_string buf " | " else vert := true;
      Buffer.add_char buf '`';
      Buffer.add_string buf l;
      if t <> CTUnit then
      begin
        Buffer.add_string buf " of ";
        compiled_type_to_buf buf t
      end) v;
    Buffer.add_string buf " ] "
| CTUnit -> Buffer.add_string buf "unit"
| CTTuple(l, _) ->
    Buffer.add_char buf '(';
    let star = ref false in
    List.iter (fun t ->
      if !star then Buffer.add_string buf " * " else star := true;
      compiled_type_to_buf buf t) l;
    Buffer.add_char buf ')'

let rec compiled_types_to_buf buf first = function
  [] -> buf
| (name, t)::tl ->
    if not first then Buffer.add_char buf '\n';
    Buffer.add_string buf ("and "^name^" = "^name^"_node located\nand "^name^"_node = ");
    compiled_type_to_buf buf t;
    Buffer.add_char buf '\n';
    compiled_types_to_buf buf false tl

let compiled_types_to_string ct =
  let buf = Buffer.create 16 in
  Buffer.add_string buf "type loc = Lexing.position * Lexing.position\ntype 'a located = { loc : loc; node : 'a }\n\n";
  Buffer.contents (compiled_types_to_buf buf true ct)

let compiled_type_to_string ct =
  let buf = Buffer.create 16 in
  compiled_type_to_buf buf ct;
  Buffer.contents buf
